"""asyncvictoria - Write data to VictoriaMetrics."""

from __future__ import annotations

from ._impl import NoCodeError as NoCodeError
from ._impl import RespError as RespError
from ._impl import RespUnknownError as RespUnknownError
from ._impl import connect as connect
from ._impl import get_data as get_data
from ._impl import get_max_ts as get_max_ts
from .model import DS as DS
from .model import Entry as Entry
from .model import EntryDelta as EntryDelta
from .sanitize import sanitize_entry as sanitize_entry
from .sanitize import sanitize_label_name as sanitize_label_name
from .sanitize import sanitize_metric_name as sanitize_metric_name
